import os as _os
import pprint as _pp
import random as _random
import typing
from argparse import ArgumentParser as _AP

import numpy as _np
import torch as _torch
import torch.distributed as _dist
import torch.multiprocessing as _mp

from pathlib import Path as _Path

import easytorch.config as _conf
import easytorch.utils as _utils
from easytorch.config.state import *
from easytorch.data import ETDataset, ETDataHandle, DiskCache as _DiskCache
from easytorch.runner import ETRunner
from easytorch.utils.logger import *
import uuid as _uuid
from datetime import datetime as _dtime
import yaml as _yaml

_sep = _os.sep
_DEFAULT_YAML = str(_Path(__file__).resolve().parent) + _sep + "config" + _sep + "default-cfg.yaml"


def _ddp_worker(rank, self, runner_cls, dataset_cls, data_handle_cls):
    self.conf['gpu'] = self.conf['gpus'][rank]
    self.conf['verbose'] = rank == MASTER_RANK
    world_size = self.conf['world_size']
    if not world_size:
        world_size = self.conf['num_gpus'] * self.conf['num_nodes']
    self.conf['world_rank'] = self.conf['node_rank'] * self.conf['num_gpus'] + rank

    self.conf['is_master'] = self.conf['world_rank'] == MASTER_RANK
    _dist.init_process_group(backend=self.conf['dist_backend'],
                             init_method=self.conf['init_method'],
                             world_size=world_size, rank=self.conf['world_rank'])
    self._run(runner_cls, dataset_cls, data_handle_cls)


def _cleanup(engine, data_handle):
    for data_handle_key in data_handle.dataloader_args:
        if data_handle.dataloader_args[data_handle_key].get('dataset') and \
                hasattr(data_handle.dataloader_args[data_handle_key]['dataset'], 'diskcache') \
                and isinstance(data_handle.dataloader_args[data_handle_key]['dataset'].diskcache, _DiskCache):
            data_handle.dataloader_args[data_handle_key]['dataset'].diskcache.clear()

    if _dist.is_initialized():
        _dist.destroy_process_group()


class EasyTorch:
    _MODES_ = [Phase.TRAIN, Phase.TEST, Phase.INFERENCE]
    _MODE_ERR_ = \
        "####  [ERROR]  ### argument 'phase' is required and must be passed to either" \
        '\n\t1). EasyTorch(..,phase=<value>,..)' \
        '\n\t2). runtime arguments 2). python main.py -ph <value> ...' \
        f'\nPossible values are:{_MODES_}'

    def __init__(self, config_source=_conf.args_parser(), dataloader_args=None, **kw):
        """
        :param config_source: argument parser, yaml file, or dict
        :param dataloader_args: as you would pass in torch.utils.Dataloader(..) class
        :param kw: Anything extra. Will override all.
        """

        self.conf = {}
        if isinstance(config_source, _AP):
            _known, _unknown = config_source.parse_known_args()
            self.conf.update(**vars(_known))
            self.conf['config_source'] = 'ArgumentParser'

        elif isinstance(config_source, dict):
            self.conf.update(**config_source)
            self.conf['config_source'] = 'dict'

        elif config_source.endswith('.yaml'):
            self.conf.update(**_yaml.safe_load(config_source))
            self.conf['config_source'] = config_source

        self.conf.update(**kw)

        self.dataloader_args = dataloader_args
        assert (self.conf.get('phase') in self._MODES_), self._MODE_ERR_

        self._device_check()
        self._ddp_setup()
        self._make_reproducible()
        self.conf.update(is_master=self.conf.get('is_master', True), world_rank=0)
        self.conf['RUN-ID'] = _dtime.now().strftime("ET-%Y%m%d-%H%M%S-") + _uuid.uuid4().hex[:4].upper()

        self.conf['save_dir'] = self.conf['output_base_dir'] + _sep + self.conf['phase'].upper() + _sep + self.conf[
            "name"]

    def _device_check(self):
        self.conf['gpus'] = self.conf['gpus'] if self.conf.get('gpus') else []
        if self.conf['verbose'] and len(self.conf['gpus']) > NUM_GPUS:
            warn(f"{len(self.conf['gpus'])} GPU(s) requested "
                 f"but {NUM_GPUS if CUDA_AVAILABLE else 'GPU(s) not'} detected. "
                 f"Using {str(NUM_GPUS) + ' GPU(s)' if CUDA_AVAILABLE else 'CPU(Much slower)'}.")
            self.conf['gpus'] = list(range(NUM_GPUS))

        if self.conf.get('world_size') and self.conf.get('dist_backend') == 'gloo':
            """Reserved gloo and world_size for CPU multi process use case."""
            self.conf['gpus'] = [None] * self.conf.get('world_size')

    def _ddp_setup(self):
        if all([self.conf['use_ddp'], len(self.conf['gpus']) >= 1]):
            self.conf['num_gpus'] = len(self.conf['gpus'])
            _os.environ['MASTER_ADDR'] = self.conf.get('master_addr', '127.0.0.1')  #
            _os.environ['MASTER_PORT'] = self.conf.get('master_port', '12355')
        else:
            self.conf['use_ddp'] = False

        """Check if want to do distributed validation"""
        self.conf['distributed_validation'] = self.conf['use_ddp'] and self.conf.get('distributed_validation', False)

    def _show_args(self):
        info(f" *** Starting phase:{self.conf['phase']}, Name:{self.conf['name']} ***", self.conf['verbose'])
        if self.conf['verbose']:
            _pp.pprint(self.conf)

    def _init_config(self, args):
        if isinstance(args, _AP):
            self.conf = vars(args.parse_args())
        elif isinstance(args, dict):
            self.conf = {**args}
        else:
            raise ValueError('2nd Argument of EasyTorch could be only one of :ArgumentParser, dict')

    def _make_reproducible(self):
        if self.conf['use_ddp'] and self.conf['seed'] is None:
            raise ValueError('Seed must be explicitly given as seed=<seed> (Eg.1, 2, 101, 102) in DDP.')

        if self.conf['seed'] is None:
            self.conf['seed'] = CURRENT_SEED

        if self.conf.get('seed_all'):
            _torch.manual_seed(self.conf['seed'])
            _torch.cuda.manual_seed_all(self.conf['seed'])
            _torch.cuda.manual_seed(self.conf['seed'])
            _np.random.seed(self.conf['seed'])
            _random.seed(self.conf['seed'])
            _torch.backends.cudnn.deterministic = True
            _torch.backends.cudnn.benchmark = False

    def _maybe_advance_run(self):
        r"""
        Checks if there already is a previous run and prompt[Y/N] so that
        we avoid accidentally overriding previous runs and lose temper.
        User can supply -f True flag to override by force.
        """
        if self.conf['force']:
            warn('Forced overriding previous logs.', self.conf['verbose'])
            return

        i = -1
        for i in range(501):
            current = self.conf['save_dir'] + f"_V{i}"
            if not _os.path.isdir(current):
                break
        self.conf['save_dir'] = self.conf['save_dir'] + f"_V{i}"
        self.conf['name'] = _Path(self.conf['save_dir']).name

    def _prepare_nn_engine(self, engine):
        engine.cache['log_header'] = 'Loss|Accuracy'
        engine.cache.update(monitor_metric='time', metric_direction='maximize')

        engine.cache[LogKey.TRAIN_LOG] = []
        engine.cache[LogKey.VALIDATION_LOG] = []
        engine.cache[LogKey.TEST_METRICS] = []

        engine.cache['best_checkpoint'] = f"best_{self.conf['name']}{CHK_EXT}"
        engine.cache['latest_checkpoint'] = f"last_{self.conf['name']}{CHK_EXT}"
        engine.cache.update(best_val_epoch=0, best_val_score=0.0)
        if engine.cache['metric_direction'] == 'minimize':
            engine.cache['best_val_score'] = MAX_SIZE
        engine.init_cache()
        engine.init_nn()

    def _run_training_and_eval(self, data_split, engine, dataset_cls):

        train_loader = engine.data_handle.get_data_loader(
            handle_key='train',
            shuffle=True,
            datset=engine.data_handle.get_dataset(Phase.TRAIN, data_split, dataset_cls),
            distributed=self.conf['use_ddp'])

        val_dataset = engine.data_handle.get_dataset(Phase.VALIDATION, data_split, dataset_cls)
        if val_dataset:
            val_loader = engine.data_handle.get_data_loader(
                handle_key='validation',
                shuffle=False,
                dataset=val_dataset,
                distributed=self.conf['use_ddp'] and self.conf.get('distributed_validation'),
                use_unpadded_sampler=True
            )
            engine.train(train_loader, val_loader)
        else:
            engine.train(train_loader, None)

        engine.save_checkpoint(engine.conf['save_dir'] + _sep + engine.cache['latest_checkpoint'])

        train_log = engine.conf['save_dir'] + _sep + ".train_log.npy"
        val_log = engine.conf['save_dir'] + _sep + ".validation_log.npy"

        _np.save(train_log, _np.array(engine.cache[LogKey.TRAIN_LOG]))
        _np.save(val_log, _np.array(engine.cache[LogKey.TRAIN_LOG]))

        engine.cache[LogKey.TRAIN_LOG] = train_log
        engine.cache[LogKey.VALIDATION_LOG] = val_log
        _utils.save_cache(self.conf, engine.cache, name=engine.conf['name'] + "_train")
        engine.cache['_saved'] = True

    def _run_test(self, data_split, engine, dataset_cls, distributed=False) -> dict:
        test_dataset = engine.data_handle.get_dataset(Phase.TEST, data_split, dataset_cls)

        dataloader = engine.data_handle.get_data_loader(
            handle_key=Phase.TEST, shuffle=False, dataset=test_dataset, distributed=distributed
        )

        best_exists = _os.path.exists(engine.conf['save_dir'] + _sep + engine.cache['best_checkpoint'])
        if best_exists and (self.conf['phase'] == Phase.TRAIN or self.conf['pretrained_path'] is None):
            engine.load_checkpoint(engine.conf['save_dir'] + _sep + engine.cache['best_checkpoint'],
                                   map_location=engine.device['gpu'], load_optimizer_state=False)

        """ Run and save experiment test scores """
        test_out = engine.evaluation(dataloader=dataloader, mode=Phase.TEST, save_predictions=True)
        test_meter = engine.reduce_scores([test_out], distributed=False)
        engine.cache[LogKey.TEST_METRICS] = [test_meter.get()]
        _utils.save_scores(self.conf['save_dir'], engine.cache, name=engine.conf['name'],
                           file_keys=[LogKey.TEST_METRICS])

        if not engine.cache.get('_saved'):
            _utils.save_cache(self.conf, engine.cache, name=f"{engine.conf['name']}_test")
        return test_out

    def _inference(self, data_split, engine, dataset_cls, distributed=False):
        infer_dataset = engine.data_handle.get_dataset(Phase.INFERENCE, data_split, dataset_cls)
        dataloader = engine.data_handle.get_data_loader(
            handle_key=Phase.INFERENCE, shuffle=False, dataset=infer_dataset, distributed=distributed,
            use_unpadded_sampler=True,
        )
        engine.inference(dataloader=dataloader)
        _utils.save_cache(self.conf, engine.cache, name=f"{engine.conf['name']}_inference")

    def run(self, runner_cls: typing.Type[ETRunner],
            dataset_cls: typing.Type[ETDataset] = ETDataset,
            data_handle_cls: typing.Type[ETDataHandle] = ETDataHandle):

        if self.conf['is_master']:
            self._maybe_advance_run()
            _os.makedirs(self.conf['save_dir'], exist_ok=self.conf['force'])

        if self.conf['verbose']:
            self._show_args()

        if self.conf.get('use_ddp'):
            _mp.spawn(_ddp_worker, nprocs=self.conf['num_gpus'],
                      args=(self, runner_cls, dataset_cls, data_handle_cls))
        else:
            self._run(runner_cls, dataset_cls, data_handle_cls)

    def _run(self, runner_cls, dataset_cls, data_handle_cls):

        engine = runner_cls(
            conf=self.conf,
            data_handle=data_handle_cls(
                conf=self.conf,
                dataloader_args=self.dataloader_args
            )
        )

        self._prepare_nn_engine(engine)

        data_split = {}
        if engine.data_handle.data_source:
            data_split = engine.data_handle.get_data_split()

        if self.conf['phase'] == Phase.TRAIN:
            self._run_training_and_eval(data_split, engine, dataset_cls)

            if self.conf['is_master'] and (data_split.get('test') or self.dataloader_args.get('test')):
                self._run_test(data_split, engine.data_handle, engine, dataset_cls)

        if self.conf['phase'] == Phase.INFERENCE:
            self._inference(data_split, engine, dataset_cls, self.conf.setdefault('distributed_inference', False))
        _cleanup(engine, engine.data_handle)



import random as _random

import cv2 as _cv2
import numpy as _np
from PIL import Image as _IMG
from skimage.util import random_noise as _noise


class RandomNoise:
    def __init__(self, p=0.5, random_crop=False, **kw):
        self.p = p
        self.noises = ['gaussian', 'localvar', 'poisson', 'salt', 'pepper', 's&p', 'speckle']
        self.amount = _np.arange(0.01, 0.08, 0.001)
        self.vars = _np.arange(0.01, 0.03, 0.001)
        self.means = _np.arange(0.0, 0.03, 0.001)
        self.random_crop = random_crop

    def _random_crop_ix(self, arr):
        h, w, _ = arr.shape
        crop_height = _random.choice(list(range(h // 10, h - 1, 5)))
        crop_width = _random.choice(list(range(w // 10, w - 1, 5)))

        max_x = arr.shape[1] - crop_width
        max_y = arr.shape[0] - crop_height

        x = _np.random.randint(0, max_x)
        y = _np.random.randint(0, max_y)
        return [y, y + crop_height, x, x + crop_width]

    def _crop_mask(self, arr, pqrs):
        p, q, r, s = pqrs
        mask = _np.zeros_like(arr)
        mask[p:q, r:s] = 255
        return

    def __call__(self, img):
        arr = _np.array(img)
        if _random.random() > self.p:
            return _IMG.fromarray(arr)

        args = {}
        noise = _random.choice(self.noises)
        if noise in ['salt', 'pepper', 's&p']:
            args['amount'] = _random.choice(self.amount)

        if noise in ['gaussian', 'speckle']:
            args['mean'] = _random.choice(self.means)
            args['var'] = _random.choice(self.vars)

            if noise == 'speckle':
                """speckle is weak"""
                args['mean'] *= 2
                args['var'] *= 2

        _arr = arr.copy()
        noisy = (_noise(_arr, noise, **args) * 255).astype(_np.uint8)
        if _random.random() <= self.p and self.random_crop:
            p, q, r, s = self._random_crop_ix(arr)
            _arr[p:q, r:s] = noisy[p:q, r:s]
        else:
            _arr = noisy

        gray = _cv2.cvtColor(_arr, _cv2.COLOR_BGR2GRAY)
        gray = _cv2.merge([gray, gray, gray])
        gray = _cv2.GaussianBlur(gray, (3, 3), 0)
        return _IMG.fromarray(gray)



import glob
import glob as _glob
import json as _json
import math as _math
import os as _os
import os.path
from os import sep as _sep
import cv2 as _cv2

import torch as _torch
import torch.distributed as _dist
import torch.utils.data as _data
from torch.utils.data import DataLoader as _DataLoader, Dataset as _Dataset
from torchvision import transforms as _tmf

import easytorch.data.datautils as _du
import easytorch.utils as _etutils
import easytorch.data.multiproc as _multi
from easytorch.utils.logger import *
import pickle as _pickle
import shutil as _shu
import uuid as _uuid
from pathlib import Path as _Path
from easytorch.config.state import *


class DiskCache:
    def __init__(self, path, verbose=True):
        self.path = path
        self.verbose = verbose

    def _gen_key(self, name):
        return _os.path.basename(name) + '-' + _uuid.uuid5(_uuid.NAMESPACE_X500, f"{name}").hex[:8]

    def add(self, name, value):
        _os.makedirs(self.path, exist_ok=True)
        key = self._gen_key(name)
        with open(self.path + _os.sep + key + ".pkl", 'wb') as file:
            _pickle.dump(value, file, _pickle.HIGHEST_PROTOCOL)
        return key

    def get(self, key):
        with open(self.path + _os.sep + key + ".pkl", 'rb') as file:
            return _pickle.load(file)

    def clear(self):
        if _os.path.exists(self.path):
            _shu.rmtree(self.path, ignore_errors=True)
            info(f"Diskcache : {self.path} cleared.", self.verbose)


class ETDataHandle:

    def __init__(self, conf: dict = None, dataloader_args: dict = None, **kw):
        self.conf = _etutils.FrozenDict(conf)
        self.dataloader_args = {'train': {}, 'validation': {}, 'test': {}, 'inference': {}}

        if dataloader_args:
            self.dataloader_args.update(**dataloader_args)

        self.diskcache = DiskCache(
            self.conf['save_dir'] + _os.sep + "_cache" + _sep + self.conf['RUN-ID'],
            self.conf['verbose']
        )
        self.data_source = conf.get('data_source')

    def get_dataset(self, handle_key, data_split, dataset_cls=None):
        if self.dataloader_args[handle_key].get('dataset'):
            return self.dataloader_args[handle_key]['dataset']

        if data_split.get(handle_key):
            dataset = dataset_cls(mode=handle_key, limit=self.conf['load_limit'], **self.conf)
            dataset.add(files=data_split[handle_key], diskcache=self.diskcache, verbose=self.conf['verbose'])
            return self.dataloader_args[handle_key].setdefault('dataset', dataset)

    def get_data_loader(self, handle_key='', distributed=False, use_unpadded_sampler=False, **kw):
        args = {**self.conf}
        args['distributed'] = distributed
        args['use_unpadded_sampler'] = use_unpadded_sampler
        args.update(self.dataloader_args.get(handle_key, {}))
        args.update(**kw)

        if args.get('dataset') is None:
            return None

        loader_args = {
            'dataset': None,
            'batch_size': 1,
            'sampler': None,
            'shuffle': False,
            'batch_sampler': None,
            'num_workers': 0,
            'pin_memory': False,
            'drop_last': False,
            'timeout': 0,
            'worker_init_fn': _multi.seed_worker if args.get('seed_all') else None
        }

        for k in loader_args.keys():
            loader_args[k] = args.get(k, loader_args.get(k))

        if loader_args['sampler'] is not None:
            """Sampler and shuffle are mutually exclusive"""
            loader_args['shuffle'] = False

        if args['distributed']:
            sampler_args = {
                'num_replicas': args.get('replicas'),
                'rank': args.get('rank'),
                'shuffle': args.get('shuffle'),
                'seed': args.get('seed')
            }

            if loader_args.get('sampler') is None:
                loader_args['shuffle'] = False  # Shuffle is mutually exclusive with sampler
                if args['use_unpadded_sampler']:
                    loader_args['sampler'] = UnPaddedDDPSampler(loader_args['dataset'], **sampler_args)
                else:
                    loader_args['sampler'] = _data.distributed.DistributedSampler(loader_args['dataset'],
                                                                                  **sampler_args)

            loader_args['num_workers'] = _multi.num_workers(args, loader_args, True)
            loader_args['batch_size'] = _multi.batch_size(args, loader_args, True)

        return _DataLoader(collate_fn=_multi.safe_collate, **loader_args)

    def get_data_split(self):
        p = _Path(self.data_source)
        split = {"test": []}
        if self.conf['phase'] == Phase.TRAIN:
            split['train'] = []
            split['validation'] = []

        if str(p).endswith('*.txt'):
            for txt in glob.glob(self.data_source):
                pth = _Path(txt)
                if pth.stem in split:
                    with open(txt) as fw:
                        split[_Path(txt).stem] = fw.read().splitlines()

        elif str(p).endswith('.json'):
            with open(self.data_source) as fw:
                split = _json.load(fw)

        else:
            if p.is_dir():
                files = _glob.glob(self.data_source + _sep + "*.*")

            elif '*' in str(p):
                files = _glob.glob(self.data_source, recursive='**' in self.data_source)

            elif p.suffix == '.txt':
                with open(str(p)) as fw:
                    files = fw.read().splitlines()
            else:
                raise ValueError(f"Unknown data source: {self.data_source}")

            files = sorted(files)
            if self.conf['phase'] == Phase.INFERENCE:
                split = {Phase.INFERENCE: files}

            else:
                split = _du.create_ratio_split(files, self.conf['split_ratio'])

        _spl = self.conf['save_dir'] + _sep + f"SPLIT_{_Path(self.conf['save_dir']).name}_{self.conf['name']}.json"
        with open(_spl, 'w') as fw:
            _json.dump(split, fw)

        for k in split:
            info(f" - Data count: {len(split[k])}, phase:{k}", self.conf['verbose'])

        return split


class ETDataset(_Dataset):
    def __init__(self, mode='init', limit=None, **kw):
        self.mode = mode
        self.limit = limit
        self.indices = []
        self.data = {}
        self.diskcache = None
        self.conf = _etutils.FrozenDict(kw)

        tmfs = [_tmf.ToTensor()]
        if kw.get('image_size'):
            tmfs.append(_tmf.Resize(tuple(kw['image_size'])))
        self.transforms = _tmf.Compose(tmfs)

    def load_index(self, file):
        r"""
        Logic to load indices of a single file.
        -Sometimes one image can have multiple indices like U-net where we have to get multiple patches of images.
        """
        self.indices.append(file)

    def _load_indices(self, files, verbose=True):
        r"""
        We load the proper indices/names(whatever is called) of the files in order to prepare minibatches.
        Only load lim numbr of files so that it is easer to debug(Default is infinite, -lim/--load-lim argument).
        """
        _files = files[:self.limit]
        _files_len = len(_files)
        if self.conf["multi_load"] and _files_len > 1:
            dataset_objs = _multi.multi_load(
                self.mode,
                _files,
                self.conf,
                self.__class__,
                diskcache=self.diskcache
            )
            self.gather(dataset_objs)
        else:
            for i, file in enumerate(_files, 1):
                info(f"Loading... {i}/{_files_len}", i % _multi.LOG_FREQ == 0)
                self.load_index(file)
        success(f'{self.mode}, {len(self)} indices Loaded.', verbose)

    def gather(self, dataset_objs):
        for d in dataset_objs:
            attributes = vars(d)
            for k, v in attributes.items():
                if isinstance(v, _etutils.FrozenDict):
                    continue

                if isinstance(v, list):
                    self.__getattribute__(f"{k}").extend(v)

                elif isinstance(attributes[f"{k}"], dict):
                    self.__getattribute__(f"{k}").update(**v)

                elif isinstance(attributes[f"{k}"], set):
                    self.__getattribute__(f"{k}").union(v)

    def __getitem__(self, index):
        file = self.indices[index]
        arr = _cv2.imread(file)
        arr = self.transforms(arr)
        return arr, index

    def __len__(self):
        return len(self.indices)

    def add(self, files, diskcache=None, verbose=True, **kw):
        r""" An extra layer for added flexibility."""
        self.diskcache = diskcache
        self._load_indices(files=files, verbose=verbose)


class UnPaddedDDPSampler(_data.Sampler):
    r"""fork from official pytorch repo: torch.data.distributed.DistributedSampler where padding is off"""
    r"""https://github.com/pytorch/"""

    r"""Sampler that restricts data loading to a subset of the dataset.

    It is especially useful in conjunction with
    :class:`torch.nn.parallel.DistributedDataParallel`. In such a case, each
    process can pass a :class`~torch.utils.data.DistributedSampler` instance as a
    :class:`~torch.utils.data.DataLoader` sampler, and load a subset of the
    original dataset that is exclusive to it.

    .. note::
        Dataset is assumed to be of constant size.

    Arguments:
        dataset: Dataset used for sampling.
        num_replicas (int, optional): Number of processes participating in
            distributed training. By default, :attr:`rank` is retrieved from the
            current distributed group.
        rank (int, optional): Rank of the current process within :attr:`num_replicas`.
            By default, :attr:`rank` is retrieved from the current distributed
            group.
        shuffle (bool, optional): If ``True`` (default), sampler will shuffle the
            indices.
        seed (int, optional): random seed used to shuffle the sampler if
            :attr:`shuffle=True`. This number should be identical across all
            processes in the distributed group. Default: ``0``.

    .. warning::
        In distributed mode, calling the :meth`set_epoch(epoch) <set_epoch>` method at
        the beginning of each epoch **before** creating the :class:`DataLoader` iterator
        is necessary to make shuffling work properly across multiple epochs. Otherwise,
        the same ordering will be always used.

    Example::

        >>> sampler = DistributedSampler(dataset) if is_distributed else None
        >>> loader = DataLoader(dataset, shuffle=(sampler is None),
        ...                     sampler=sampler)
        >>> for epoch in range(start_epoch, n_epochs):
        ...     if is_distributed:
        ...         sampler.set_epoch(epoch)
        ...     train(loader)
    """

    def __init__(self, dataset, num_replicas=None, rank=None, shuffle=True, seed=0):
        if num_replicas is None:
            if not _dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            num_replicas = _dist.get_world_size()
        if rank is None:
            if not _dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = _dist.get_rank()
        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = 0

        """For unpadded sampling"""
        self.num_samples = int(_math.ceil((len(self.dataset) - self.rank) * 1.0 / self.num_replicas))
        self.total_size = len(self.dataset)

        self.shuffle = shuffle
        self.seed = seed

    def __iter__(self):
        if self.shuffle:
            # deterministically shuffle based on epoch and seed
            g = _torch.Generator()
            g.manual_seed(self.seed + self.epoch)
            indices = _torch.randperm(len(self.dataset), generator=g).tolist()
        else:
            indices = list(range(len(self.dataset)))

        """Do not pad anything"""
        # add extra samples to make it evenly divisible
        # indices += indices[:(self.total_size - len(indices))]

        assert len(indices) == self.total_size

        # subsample
        indices = indices[self.rank:self.total_size:self.num_replicas]
        assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self):
        return self.num_samples

    def set_epoch(self, epoch):
        r"""
        Sets the epoch for this sampler. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.

        Arguments:
            epoch (int): Epoch number.
        """
        self.epoch = epoch



import multiprocessing as _mp
import traceback as _tb
from functools import partial as _partial
from typing import Callable

import numpy as _np
import torch as _torch
from torch.utils.data._utils.collate import default_collate as _default_collate

LOG_FREQ = 100


def _job(total, func, i, f):
    print(f"Working on: [ {i}/{total} ]", end='\n' if i % LOG_FREQ == 0 else '\r')
    try:
        return func(f)
    except Exception as e:
        _tb.print_exc()
        print(f"{f} ### {e}")


def multiRun(nproc: int, data_list: list, func: Callable) -> list:
    _files = []
    for ix, file in enumerate(data_list, 1):
        _files.append([ix, file])

    with _mp.Pool(processes=nproc) as pool:
        return list(
            pool.starmap(_partial(_job, len(_files), func), _files)
        )


def safe_collate(batch):
    r"""Safely select batches/skip dataset_cls(errors in file loading."""
    return _default_collate([b for b in batch if b])


def num_workers(conf, loader_args, distributed=False):
    if distributed:
        return (loader_args['num_workers'] + conf['num_gpus'] - 1) // conf['num_gpus']
    return loader_args.get('num_workers', 0)


def batch_size(conf, loader_args, distributed=False):
    if distributed:
        loader_args['batch_size'] = loader_args['batch_size'] // conf['num_gpus']
    return loader_args['batch_size']


def seed_worker(worker_id):
    seed = (int(_torch.initial_seed()) + worker_id) % (2 ** 32 - 1)
    _np.random.seed(seed)


def _et_data_job(file, mode, conf, dataset_cls, diskcache):
    dataset = dataset_cls(mode=mode, **conf)
    try:
        dataset.add(files=[file], diskcache=diskcache, verbose=False)
    except Exception as e:
        _tb.print_exc()
        print(f"{file} ### {e}")
    return dataset


def et_data_job(mode, conf, dataset_cls, total, verbose, diskcache, i, file):
    if verbose:
        print(f"Working on: [ {i} / {total} ]", end='\n' if i % LOG_FREQ == 0 else '\r')
    return _et_data_job(file, mode, conf, dataset_cls, diskcache)


def multi_load(mode, files, conf, dataset_cls, diskcache=None) -> list:
    r"""Note: Only works with easytorch's default args from easytorch import args"""
    _files = []
    for ix, f in enumerate(files, 1):
        _files.append([ix, f])

    if conf.get("multi_load") and len(files) > 1:
        nw = min(num_workers(conf, conf, conf.get('use_ddp')), len(_files))
        with _mp.Pool(processes=max(1, nw)) as pool:
            dataset_list = list(
                pool.starmap(
                    _partial(
                        et_data_job, mode, conf, dataset_cls, len(_files), conf.get('verbose'), diskcache
                    ),
                    _files
                )
            )
            return [d for d in dataset_list if len(d) >= 1]

    dataset_list, total = [], len(_files)
    for ix, file in _files:
        if conf.get('verbose'):
            print(f"Loading... {ix}/{total}", end='\n' if ix % LOG_FREQ == 0 else '\r')
        dataset_list.append(_et_data_job(file, mode, conf, dataset_cls, diskcache))

    return [d for d in dataset_list if len(d) >= 1]
